#include <stdio.h>
#include <stdlib.h>
#include <openssl/buffer.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <string.h>

#define SHA1_WITH_RSA "007"
#define NEW_SHA1_WITH_RSA "008"
#define CERT "-----BEGIN CERTIFICATE-----"
#define CERT_REQ "-----BEGIN CERTIFICATE REQUEST-----"
#define DEBUG(str)
// printf("%s\n",str);
#define MARKER "\012#\012"

int cert_type = 0;

static int
slurp(FILE* fd, char** pb)
{
  int n = 0;
  int nn;
  if (!fd) return -1;
  if (!pb) return -1;
  if ((*pb = malloc(4097))==NULL) return -1;

  for(;;) {
    nn = fread((*pb)+n, 1, 4096, fd);
    if (nn <= 0) break;
    n+=nn;
    if ((*pb = realloc(*pb, n+4097))==NULL) return -1;
  }
  if (nn<0) return -1;
  (*pb)[n] = '\0';
  if ((*pb = realloc(*pb, n+1))==NULL) return -1;
  return n;
}

BIO *set_read_BIO_from_buf(char* buf, int len)
{
  BIO* rbio;
  BUF_MEM* bm;
  if (!(rbio = BIO_new(BIO_s_mem()))) goto srb_err;
  if (!(bm = BUF_MEM_new()))  goto srb_err;
  if (!BUF_MEM_grow(bm, len)) goto srb_err;
  memcpy(bm->data, buf, len);
  BIO_set_mem_buf(rbio, bm, 0 /*not used*/);
  return rbio;
srb_err:
  return NULL;
}

X509 *read_X509_data(BIO *bio) 
{
    if ( cert_type == 1 ) {
	return PEM_read_bio_X509(bio,NULL,NULL,NULL);
    } else {
	return PEM_read_bio_X509_REQ(bio,NULL,NULL,NULL);
    };
}

EVP_PKEY *extract_pubkey(X509 *x509)
{
    if ( cert_type == 1 ) {
        return (EVP_PKEY*)X509_extract_key(x509);
    } else {
        return (EVP_PKEY*)X509_REQ_extract_key((X509_REQ*)x509);
    };
}


X509 *extract_certificate(char* cert)
{
  X509* x509;
  BIO* rbio;
  if (!(rbio = set_read_BIO_from_buf(cert, strlen(cert)))) goto ec_err;
  if (!(x509=read_X509_data(rbio))) goto ec_err;
  BIO_free(rbio);
  return x509;

ec_err:
  return NULL;
}

int do_rsa ( char *data ,int data_len,char *sign,int sign_len,char *key,int key_len )
{
    EVP_PKEY *pkey;   // Public key
    EVP_MD_CTX     md_ctx;  //Digest kontekst
    X509 *x509;       //Sertifikaat

    char ub64[4096];  // viit base64-dekodeeritud hashile
    int ub64_len;  // base64-dekodeeritud hashi pikkus.

    if ( (x509 = extract_certificate(key)) == NULL ) {
        printf("TECH\n");
        exit(30);
    };
    
    if ( (pkey = extract_pubkey(x509)) == NULL) {
        printf("TECH\n");
        exit (31);
    }

    if ((ub64_len = EVP_DecodeBlock(ub64,sign,sign_len) ) < 1 ) {
        printf("TECH\n");
        exit(32);
    };

    EVP_VerifyInit(&md_ctx, EVP_sha1());
    EVP_VerifyUpdate(&md_ctx, data, data_len);

    if ( EVP_VerifyFinal(&md_ctx, ub64, EVP_PKEY_size(pkey), pkey) != 1 ) {
        printf("NOK\n");
    } else {
        printf("OK\n");
    };

    if ( cert_type == 1) {
	X509_free(x509);
    } else {
	X509_REQ_free((X509_REQ*)x509);
    };

    EVP_PKEY_free (pkey);

    DEBUG("SHA1_WITH_RSA");
    return 0;
}

int main(int argc,char **argv)
{

    int file_len;
    char *text;
    
    // Mrab jrgnevate andmete kodeeringu -
    // 001 : allkiri on hash andmetest MD5 kodeeringus
    // 002 : allkiri on hash andmetest SHA1 kodeeringus
    // 007 : allkiri on hash andmetest MD5 kodeeringus, mis on krptitud RSA'ga
    char *type;
    
    // Lahtise tekstiga andmed, millele on 001 ja 002 puhul lisatud mingi string.
    char *data;
    int data_len;

    // Base64-kodeeritud hash, mis on vastavalt tbile type -
    // 001 : MD5 digest data'st
    // 002 : SHA1 digest data'st
    // 007 : MD5 digest datast krpteeritud RSA salajase vtmega.
    char *sign;
    int sign_len;

    // 'Vti', mida interpreteeritakse vastavalt tbie type -
    //
    char *key;
    int key_len;

    char *tmp;
    // Imeme faili sisse.
    if ( ((file_len = slurp(stdin, &text)) < 0 )) {
        // Viga sisselugemisel, minema.
        printf("TECH\n");
        exit (1);
    };
                          
    // Leiame komponentide asukoha - eraldatud on nad stringiga "\n#\n";

    
    // Selgitame vlja, mismoodi jrgnevaid andmeid interpreteerida.
    // Et mitte if-then rgastikke tekitada, kasutame eraldi funktsioone.

    type = text;

    if ( strncmp( type, SHA1_WITH_RSA, strlen(SHA1_WITH_RSA)) &&
         strncmp( type, NEW_SHA1_WITH_RSA, strlen(NEW_SHA1_WITH_RSA))) {
        printf("TECH\n");
        exit (2);
    };
 
    tmp = text;

    if ( (tmp = strstr(tmp,MARKER)) == NULL) {
        // Ei suuda leida data algust.
        DEBUG("Pole datat?")
        printf("TECH\n");
        exit(3);
    };
    //                          __ tmp
    //                          |
    data = (tmp += 3);   // "fsdfs\n#\nDATA ALGAB SIIN...
    //                           +3=|_ data,tmp
    if ( (tmp = strstr(tmp,MARKER)) == NULL) {
        // Ei suuda leida signatuuri algust.
        DEBUG("Pole signatuuri?")
        printf("TECH\n");
        exit(4);
    };
    data_len = tmp - data;
    sign = (tmp += 3);
    
    if ( (tmp = strstr(tmp,MARKER)) == NULL) {
        // Ei suuda leida sertifikaadi* algust.
        DEBUG("Pole vtit?")
        printf("TECH\n");
        exit(5);
    };
    sign_len = tmp - sign;
    key = (tmp +=3 );

    if ( (tmp = strstr(tmp,MARKER)) == NULL) {
        // Ei suuda leida lo~ppu?.
        DEBUG("Lopp puudu?")
        printf("TECH\n");
        exit(6);
    };
    key_len = tmp - key;

    // Selgitame vlja, kas avalik vti peitub sertifikaadis(a'la Eesti hispank) vi 
    // sertifikaaditaotluses (a'la Hansapank) 
    
    if ( strncmp(key,CERT,strlen(CERT)) == 0 ) { //Sertifikaat 
	cert_type = 1;
    };
    if ( strncmp(key,CERT_REQ,strlen(CERT_REQ)) == 0 ) { //Sertifikaaditaotlus 
	cert_type = 2;
    };
    if ( ! cert_type ) {  // Polnud sobiva pisega sert
	printf("TECH");
        exit(7);
    };
     
    do_rsa(data,data_len,sign,sign_len,key,key_len);

    return 0;
}
