#include <stdio.h>
#include <stdlib.h>
#include <openssl/buffer.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <string.h>


#define SHA1_WITH_RSA "007"
#define NEW_SHA1_WITH_RSA "008"

#define DEBUG(str)
//printf("%s\n",str);
#define MARKER "\012#\012"


static int
slurp(FILE* fd, char** pb)
{
  int n = 0;
  int nn;
  if (!fd) return -1;
  if (!pb) return -1;
  if ((*pb = malloc(4097))==NULL) return -1;

  for(;;) {
    nn = fread((*pb)+n, 1, 4096, fd);
    if (nn <= 0) break;
    n+=nn;
    if ((*pb = realloc(*pb, n+4097))==NULL) return -1;
  }
  if (nn<0) return -1;

  /* add NULL termination and shrink the buffer */

  (*pb)[n] = '\0';
  if ((*pb = realloc(*pb, n+1))==NULL) return -1;
  return n;
}


BIO *set_read_BIO_from_buf(char* buf, int len)
{
  BIO* rbio;
  BUF_MEM* bm;
  if (!(rbio = BIO_new(BIO_s_mem()))) goto srb_err;
  if (!(bm = BUF_MEM_new()))  goto srb_err;
  if (!BUF_MEM_grow(bm, len)) goto srb_err;
  memcpy(bm->data, buf, len);
  BIO_set_mem_buf(rbio, bm, 0 /*not used*/);
  return rbio;
srb_err:
  return NULL;
}

EVP_PKEY* open_private_key(char* privatekey, char* password)
{
  EVP_PKEY* pkey;
  BIO* rbio;
  if (!(rbio = set_read_BIO_from_buf(privatekey, strlen(privatekey)))) goto opk_err;
  if (!(pkey=PEM_read_bio_PrivateKey(rbio,NULL,NULL,(void*)password))) goto opk_err;
  BIO_free(rbio);
  return pkey;

opk_err:
  return NULL;
}

void pm(unsigned char *md, int len)
{
        int i;

        for (i=0; i<len; i++)
                printf("%c",md[i]);
}

int do_rsa ( char *data ,int data_len,  char *key, int key_len, char *pwd, int pwd_len)
{
    EVP_PKEY *pkey;
    EVP_MD_CTX     md_ctx;

    char b64[4096];  // viit base64-dekodeeritud hashile
    int b64_len;  // base64-dekodeeritud hashi pikkus.
    int sig_len;
    char *md;
    char *pwd_string;
    
    if ((pwd_string = malloc(pwd_len+1)) == NULL ) {
        printf("TECH\n");
        exit(25);
    };
    strncpy(pwd_string,pwd,pwd_len);
    *(pwd_string+pwd_len)='\000';

    if ( (pkey = open_private_key(key,pwd_string)) == NULL ) {
        printf("TECH\n");
        exit(30);    // Psufraas vale/puudu?
    };
    
    sig_len = EVP_PKEY_size(pkey);
    
    if ((md = malloc(sig_len)) == NULL ) {
        printf("TECH\n");
        exit(35);
    };

    EVP_SignInit(&md_ctx, EVP_sha1());
    EVP_SignUpdate (&md_ctx, data, data_len);
    
    if ( EVP_SignFinal(&md_ctx, md, &sig_len, pkey) !=1 ) {
        printf("TECH");
        exit(36);
    };
    
    EVP_PKEY_free (pkey);

    if ((b64_len = EVP_EncodeBlock(b64, md, sig_len)) < 1 ) {
        printf("TECH\n");
        exit(20);
    };

    pm(NEW_SHA1_WITH_RSA,strlen(NEW_SHA1_WITH_RSA));
    printf(MARKER);
    pm(data,data_len);
    printf(MARKER);
    pm(b64,b64_len);
    printf(MARKER);

    DEBUG("SHA1_WITH_RSA");
    return 0;
}


int main(int argc,char **argv)
{

    char *text;
    int file_len;

    // 007 : allkiri on hash andmetest SHA1 kodeeringus, mis on krptitud RSA'ga
    char *type;
    
    // Lahtise tekstiga andmed.
    char *data;
    int data_len;

    // 'Vti', tegelikult RSA privaatvti PEM-kodeeringus
    //
    char *key;
    int key_len;

    // RSA privaatvtme psufraas
    char *pwd;
    int pwd_len;
    

    char *tmp;
    // Imeme faili sisse.
    if ( ((file_len = slurp(stdin, &text)) < 0 )) {
        // Viga sisselugemisel, minema.
        printf("TECH\n");
        exit (1);
    };
    // Leiame komponentide asukoha - eraldatud on nad stringiga "\n#\n";

    
    // Selgitame vlja, mismoodi jrgnevaid andmeid interpreteerida.
    // Et mitte if-then rgastikke tekitada, kasutame eraldi funktsioone.

    type = text;

    if ( strncmp( type, SHA1_WITH_RSA, strlen(SHA1_WITH_RSA)) &&
         strncmp( type, NEW_SHA1_WITH_RSA, strlen(NEW_SHA1_WITH_RSA))) {
        printf("TECH\n");
        exit (2);
    };
    
    tmp = text;

    if ( (tmp = strstr(tmp,MARKER)) == NULL) {
        // Ei suuda leida data algust.
        DEBUG("Pole datat?")
        printf("TECH\n");
        exit(3);
    };
    //                            __ tmp
    //                            |
    data = (tmp += 3);   // "fsdfs\n#\nDATA ALGAB SIIN...
    //                              +3=|_ data,tmp
    if ( (tmp = strstr(tmp,MARKER)) == NULL) {
        // Ei suuda leida vtme algust.
        DEBUG("Pole vtit?")
        printf("TECH\n");
        exit(4);
    };
    data_len = tmp - data;
    key = (tmp +=3 );

    if ( (tmp = strstr(tmp,MARKER)) == NULL) {
        // Ei suuda leida  psufraasi algust.
        DEBUG("Psufraas puudu?")
        printf("TECH\n");
        exit(5);
    };
    key_len = tmp - key;

    pwd = (tmp +=3 );
    if ( (tmp = strstr(tmp,MARKER)) == NULL) {
        // Ka lpus peab olema \n#\n
        DEBUG("Lpp puudu?")
        printf("TECH\n");
        exit(6);
    };
    pwd_len = tmp - pwd;
    

    OpenSSL_add_all_algorithms();  // Magic word :)
    do_rsa(data,data_len,key,key_len,pwd,pwd_len);

    return 0;
}
